import json
import os
import subprocess
import boto3
import re
import time
from tibanna.awsem import (
    AwsemRunJson,
    AwsemPostRunJson,
    AwsemPostRunJsonOutput
)
from tibanna.nnested_array import (
    run_on_nested_arrays2,
    flatten,
    create_dim
)
from .target import Target, SecondaryTargetList
from . import log


downloadlist_filename = "download_command_list.txt"
mountlist_filename = "mount_command_list.txt"
input_yml_filename = "inputs.yml"
env_filename = "env_command_list.txt"
INPUT_DIR = "/data1/input"  # data are downloaded to this directory
INPUT_MOUNT_DIR_PREFIX = "/data1/input-mounted-"  # data are mounted to this directory + bucket name


def decode_run_json(input_json_file, kms_key_id=None):
    """reads a run json file and creates three text files:
    download command list file (commands to download input files from s3)
    input yml file (for cwl/wdl/snakemake run)
    env list file (environment variables to be sourced)
    """
    # read json file
    with open(input_json_file, 'r') as f:
        runjson = AwsemRunJson(**json.load(f))
    runjson_input = runjson.Job.Input
    language = runjson.Job.App.language

    # create a download command list file from the information in json
    create_download_command_list(downloadlist_filename, runjson_input)

    # create a bucket-mounting command list file
    create_mount_command_list(mountlist_filename, runjson_input, kms_key_id=kms_key_id)

    # create an input yml file to be used on awsem
    if language in ['wdl', 'wdl_v1', 'wdl_draft2']:  # wdl
        create_input_for_wdl(input_yml_filename, runjson_input)
    elif language == 'snakemake':  # snakemake
        create_input_for_snakemake(input_yml_filename, runjson_input)
    else:  # cwl
        create_input_for_cwl(input_yml_filename, runjson_input)

    # create a file that defines environmental variables
    create_env_def_file(env_filename, runjson, language)


def create_mount_command_list(mountlist_filename, runjson_input,
                              kms_key_id=None):
    """ This function creates a mount point directory and starts goofys
        with some default arguments.
        Note that KMS key arguments are needed for the mount if encryption
        is enabled.
    """
    buckets_to_be_mounted = set()
    for category in ["Input_files_data", "Secondary_files_data"]:
        for inkey, v in getattr(runjson_input, category).items():
            if v.mount:
                buckets_to_be_mounted.add(v.dir_)
    with open(mountlist_filename, 'w') as f:
        for b in sorted(buckets_to_be_mounted):
            f.write("mkdir -p %s\n" % (INPUT_MOUNT_DIR_PREFIX + b))
            if kms_key_id:
                f.write("goofys --sse-kms %s -f %s %s &\n" % (kms_key_id, b, INPUT_MOUNT_DIR_PREFIX + b))
            else:
                f.write("goofys -f %s %s &\n" % (b, INPUT_MOUNT_DIR_PREFIX + b))


def create_download_command_list(downloadlist_filename, runjson_input):
    """create a download command list file from the information in json"""
    with open(downloadlist_filename, 'w') as f:
        for category in ["Input_files_data", "Secondary_files_data"]:
            for inkey, v in getattr(runjson_input, category).items():
                if v.mount:  # do not download if it will be mounted
                    continue
                if inkey.startswith('file://'):
                    target = inkey.replace('file://', '')
                    print("key %s will be downloaded to target %s" % (v.path, inkey))
                    run_on_nested_arrays2(v.path, target, add_download_cmd, data_bucket=v.dir_,
                                          profile=v.profile, f=f, unzip=v.unzip)
                else:
                    target_template = INPUT_DIR + "/%s"
                    if not v.rename or len(flatten(v.rename)) == 0:
                        rename = create_dim(v.path, empty=True)
                    else:
                        rename = v.rename
                    run_on_nested_arrays2(v.path, rename, add_download_cmd, data_bucket=v.dir_,
                                          profile=v.profile, f=f, unzip=v.unzip, target_template=target_template)


def add_download_cmd(data_file, rename, data_bucket, profile, f, unzip, target_template='%s'):
    if data_file:
        if not rename:
            rename = data_file
        target = target_template % rename
        cmd = create_download_cmd(data_bucket, data_file, target, profile, unzip)
        f.write(cmd + '\n')


def determine_key_type(bucket, key, profile):
    """Return values : 'File', 'Folder' or 'Does not exist'"""
    if profile:
        s3 = boto3.session(profile_name=profile).client('s3')
    else:
        s3 = boto3.client('s3')
    if not key:
        raise Exception("Cannot determine key type - no key is specified")
    if not bucket:
        raise Exception("Cannot determine key type - no bucket is specified")
    if key.endswith('/'):
        key = key.rstrip('/')
    res = s3.list_objects_v2(Bucket=bucket, Prefix=key + '/')
    if not 'KeyCount' in res:
        raise Exception("Cannot determine key type - no response from S3")
    if res['KeyCount'] == 0:
        res2 = s3.list_objects_v2(Bucket=bucket, Prefix=key)
        if not 'KeyCount' in res2:
            raise Exception("Cannot determine key type - no response from S3")
        elif res2['KeyCount'] == 0:
            return 'Does not exist'  # key does not exist
        # The file itself may be a prefix of another file (e.v. abc.vcf.gz vs abc.vcf.gz.tbi)
        # but it doesn't matter.
        else:
            return 'File'
    else:
        # data_file is a folder
        return 'Folder'


def create_download_cmd(data_bucket, data_file, target, profile, unzip=''):
    profile_flag = ' --profile ' + profile if profile else ''
    format_list = [data_bucket, data_file, target, profile_flag]
    key_type = determine_key_type(data_bucket, data_file, profile)
    if key_type == 'Does not exist':
        raise Exception("Cannot download file s3://%s/%s - file does not exist." % (data_bucket, data_file))
    elif key_type == 'File':
        download_cmd = 'aws s3 cp s3://{0}/{1} {2}{3}'.format(*format_list)
        if unzip == 'gz':
            unzip_cmd = 'gunzip {2}'
        elif unzip == 'bz2':
            unzip_cmd = 'bzip2 -d {2}'
        else:
            unzip_cmd = ''
        cmd = download_cmd + '; ' + unzip_cmd
        return cmd.format(*format_list)
    else: # key_type == 'Folder':
        download_cmd = 'aws s3 cp --recursive s3://{0}/{1} {2}{3}'.format(*format_list)
        if unzip == 'gz':
            unzip_cmd = 'for f in `find {2} -type f`; do if [[ $f =~ \\.gz$ ]]; then gunzip $f; fi; done;'
        elif unzip == 'bz2':
            unzip_cmd = 'for f in `find {2} -type f`; do if [[ $f =~ \\.bz2$ ]]; then bzip2 -d $f; fi; done;'
        else:
            unzip_cmd = ''
        cmd = download_cmd + '; ' + unzip_cmd
        return cmd.format(*format_list)


# create an input yml file for cwl-runner
def create_input_for_cwl(input_yml_filename, runjson_input):
    yml = runjson_input.as_dict_as_cwl_input(INPUT_DIR, INPUT_MOUNT_DIR_PREFIX)
    with open(input_yml_filename, 'w') as f_yml:
        json.dump(yml, f_yml, indent=4, sort_keys=True)


def create_input_for_wdl(input_yml_filename, runjson_input):
    yml = runjson_input.as_dict_as_wdl_input(INPUT_DIR, INPUT_MOUNT_DIR_PREFIX)
    with open(input_yml_filename, 'w') as f_yml:
        json.dump(yml, f_yml, indent=4, sort_keys=True)


def create_input_for_snakemake(input_yml_filename, runjson_input):
    pass  # for now assume no input yml


# create a file that defines environmental variables
def create_env_def_file(env_filename, runjson, language):
    # I have to use these variables after this script finishes running.
    # I didn't use os.environ + os.system('bash') because that would remove the other
    # env variables set before this script started running.
    app = runjson.Job.App
    with open(env_filename, 'w') as f_env:
        f_env.write("export LANGUAGE={}\n".format(app.language))
        if language in ['wdl', 'wdl_v1', 'wdl_draft2']:
            f_env.write("export WDL_URL={}\n".format(app.wdl_url))
            f_env.write("export MAIN_WDL={}\n".format(app.main_wdl))
            f_env.write("export WORKFLOW_ENGINE={}\n".format(app.workflow_engine))
            f_env.write("export RUN_ARGS={}\n".format(app.run_args))
            f_env.write("export WDL_FILES=\"{}\"\n".format(' '.join(app.other_wdl_files.split(','))))
        elif language == 'snakemake':
            f_env.write("export SNAKEMAKE_URL={}\n".format(app.snakemake_url))
            f_env.write("export MAIN_SNAKEMAKE={}\n".format(app.main_snakemake))
            f_env.write("export SNAKEMAKE_FILES=\"{}\"\n".format(' '.join(app.other_snakemake_files.split(','))))
            f_env.write("export COMMAND=\"{}\"\n".format(app.command.replace("\"", "\\\"")))
            f_env.write("export CONTAINER_IMAGE={}\n".format(app.container_image))
        elif language == 'shell':
            f_env.write("export COMMAND=\"{}\"\n".format(app.command.replace("\"", "\\\"")))
            f_env.write("export CONTAINER_IMAGE={}\n".format(app.container_image))
        else:  # cwl
            f_env.write("export CWL_URL={}\n".format(app.cwl_url))
            f_env.write("export MAIN_CWL={}\n".format(app.main_cwl))
            f_env.write("export CWL_FILES=\"{}\"\n".format(' '.join(app.other_cwl_files.split(','))))
            f_env.write("export RUN_ARGS={}\n".format(app.run_args))
        # other env variables
        env_preserv_str = ''
        docker_env_str = ''
        if runjson.Job.Input.Env:
            for ev, val in sorted(runjson.Job.Input.Env.items()):
                f_env.write("export {}={}\n".format(ev, val))
                env_preserv_str = env_preserv_str + "--preserve-environment " + ev + " "
                docker_env_str = docker_env_str + "-e " + ev + " "
        f_env.write("export PRESERVED_ENV_OPTION=\"{}\"\n".format(env_preserv_str))
        f_env.write("export DOCKER_ENV_OPTION=\"{}\"\n".format(docker_env_str))


def download_workflow():
    language = os.environ.get('LANGUAGE')
    if language == 'shell':
        return
    local_wfdir = os.environ.get('LOCAL_WFDIR')
    subprocess.call(['mkdir', '-p', local_wfdir])

    if language in ['wdl', 'wdl_v1', 'wdl_draft2']:
        main_wf = os.environ.get('MAIN_WDL', '')
        wf_files = os.environ.get('WDL_FILES', '')
        wf_url = os.environ.get('WDL_URL')
    elif language == 'snakemake':
        main_wf = os.environ.get('MAIN_SNAKEMAKE', '')
        wf_files = os.environ.get('SNAKEMAKE_FILES', '')
        wf_url = os.environ.get('SNAKEMAKE_URL')
    else:
        main_wf = os.environ.get('MAIN_CWL', '')
        wf_files = os.environ.get('CWL_FILES', '')
        wf_url = os.environ.get('CWL_URL')
    # turn into a list
    if not wf_files:
        wf_files = []
    elif ' ' in wf_files:
        wf_files = wf_files.split(' ')
    else:
        wf_files = [wf_files]
    wf_files.append(main_wf)
    wf_url = wf_url.rstrip('/')

    print("main workflow file: %s" % main_wf)
    print("workflow files: " + str(wf_files))

    s3 = boto3.client('s3')
    for wf_file in wf_files:
        target = "%s/%s" % (local_wfdir, wf_file)
        source = "%s/%s" % (wf_url, wf_file)

        if wf_url.startswith('http'):
            print("downloading via wget (public file)")
            subprocess.call(["wget", "-O" + target, source])
        elif wf_url.startswith('s3'):
            wf_loc = wf_url.replace('s3://', '')
            bucket_name = wf_loc.split('/')[0]
            if len(wf_loc.split('/')) > 1:
                subdirectory = wf_loc.replace(bucket_name + '/', '')
                key = subdirectory + '/' + wf_file
            else:
                key = wf_file
            print("downloading key %s from bucket %s to target %s" % (key, bucket_name, target))
            if '/' in target:
                targetdir = re.sub('[^/]+$', '', target)
                subprocess.call(["mkdir", "-p", targetdir])
            s3.download_file(Bucket=bucket_name, Key=key, Filename=target)


def read_md5file(md5file):
    with open(md5file, 'r') as md5_f:
        md5dict = dict()
        for line in md5_f:
            a = line.split()
            path = a[1]
            md5sum = a[0]
            md5dict[path] = md5sum
    return md5dict


def create_output_files_dict(language='cwl', execution_metadata=None, md5dict=None, strict=True):
    """create a dictionary that contains 'path', 'secondaryFiles', 'md5sum' with argnames as keys.
    For snakemake and shell, returns an empty dictionary (execution_metadata not required).
    secondaryFiles is added only if the language is cwl.
    execution_metadata is a dictionary read from wdl/cwl execution log json file.
    md5dict is a dictionary with key=file path, value=md5sum (optional).
    if strict is set False, then it does not check executio_metadata exists for cwl/wdl."""
    if language in ['cwl', 'cwl_v1', 'wdl', 'wdl_v1', 'wdl_draft2'] and not execution_metadata and strict:
        raise Exception("execution_metadata is required for cwl/wdl.")
    out_meta = dict()
    if language in ['wdl', 'wdl_v1', 'wdl_draft2']:
        for argname, outfile in execution_metadata['outputs'].items():
            if outfile:
                out_meta[argname] = {'path': outfile}
    elif language == 'snakemake' or language == 'shell':
        out_meta = {}
    else:  # cwl, cwl_v1
        # read cwl output json file
        out_meta = execution_metadata

    # add md5
    if not md5dict:
        md5dict = {}
    for of, ofv in out_meta.items():
        if ofv['path'] in md5dict:
            ofv['md5sum'] = md5dict[ofv['path']]
        if 'secondaryFiles' in ofv:
            for sf in ofv['secondaryFiles']:
                if sf['path'] in md5dict:
                    sf['md5sum'] = md5dict[sf['path']]

    return out_meta


def read_postrun_json(jsonfile):
    # read old json file
    with open(jsonfile, 'r') as json_f:
        prj = AwsemPostRunJson(**json.load(json_f))
    return prj


def format_postrun_json(prj):
    return json.dumps(prj.as_dict(), indent=4, sort_keys=True)


def write_postrun_json(jsonfile, prj):
    with open(jsonfile, 'w') as f:
        f.write(format_postrun_json(prj))


def update_postrun_json_init(json_old, json_new):
    """Update postrun json with just instance ID, filesystem and availability zone"""
    # read old json file
    prj = read_postrun_json(json_old)

    # simply add instance ID, file system and availability zone
    prj.Job.instance_id = os.getenv('INSTANCE_ID')
    prj.Job.filesystem = os.getenv('EBS_DEVICE')
    prj.Job.instance_availablity_zone = os.getenv('INSTANCE_AVAILABILITY_ZONE')
    prj.Job.instance_type = os.getenv('INSTANCE_TYPE')

    # write to new json file
    write_postrun_json(json_new, prj)


def update_postrun_json_upload_output(json_old, execution_metadata_file, md5file, json_new,
                                      language='cwl_v1', strict=True, upload=True, endpoint_url=None):
    """Update postrun json with output files.
    if strict is set false, it does not check execution metadata is required for cwl/wdl."""
    # read old json file and prepare postrunjson skeleton
    prj = read_postrun_json(json_old)

    # read md5 file
    md5dict = read_md5file(md5file)

    # read execution metadata file
    if execution_metadata_file:
        with open(execution_metadata_file, 'r') as f:
            execution_metadata = json.load(f)
    else:
        execution_metadata = None
    output_files = create_output_files_dict(language, execution_metadata, md5dict, strict=strict)

    # create output files for postrun json
    prj.Job.Output.add_output_files(output_files)

    # upload output to S3 (this also updates postrun json)
    if upload:
        upload_output(prj, endpoint_url=endpoint_url)

    # write to new json file
    write_postrun_json(json_new, prj)


def upload_output(prj, endpoint_url=None):
    # parsing output_target and uploading output files to output target
    upload_to_output_target(prj.Job.Output, prj.config.encrypt_s3_upload,
                            kms_key_id=prj.config.kms_key_id, endpoint_url=endpoint_url)


def upload_to_output_target(prj_out, encrypt_s3_upload=False, kms_key_id=None, endpoint_url=None):
    # parsing output_target and uploading output files to output target
    output_bucket = prj_out.output_bucket_directory
    output_argnames = prj_out.output_files.keys()
    output_target = prj_out.alt_output_target(output_argnames)

    for k in output_target:
        target = Target(output_bucket)

        # 'file://' output targets
        if target.is_custom_target(k):
            print("processing custom (path-based) target %s" % k)
            target.parse_custom_target(k, output_target[k])
            if target.is_valid:
                print("Target is valid. Uploading..")
                target.upload_to_s3(encrypt_s3_upload=encrypt_s3_upload, endpoint_url=endpoint_url)
            else:
                raise Exception("Invalid target %s -> %s: failed to upload" % k, output_target[k])
        else:
            # legitimate CWL/WDL output targets
            print("processing argument-based target %s" % k)
            target.parse_cwl_target(k, output_target.get(k, ''), prj_out.output_files)
            if target.is_valid:
                print("Target is valid. Uploading..")
                target.upload_to_s3(encrypt_s3_upload=encrypt_s3_upload,
                                    kms_key_id=kms_key_id, endpoint_url=endpoint_url)
                prj_out.output_files[k].add_target(target.dest)

                # upload secondary files
                secondary_output_files = prj_out.output_files[k].secondaryFiles
                if secondary_output_files:
                    stlist = SecondaryTargetList(output_bucket)
                    stlist.parse_target_values(prj_out.secondary_output_target.get(k, []))
                    stlist.reorder_by_source([sf.path for sf in secondary_output_files])
                    for st in stlist.secondary_targets:
                        st.upload_to_s3(encrypt_s3_upload=encrypt_s3_upload,
                                        kms_key_id=kms_key_id, endpoint_url=endpoint_url)
                    for i, sf in enumerate(secondary_output_files):
                        sf.add_target(stlist.secondary_targets[i].dest)
            else:
                raise Exception("Failed to upload to output target %s" % k)


def save_total_sizes():
    os.environ['INPUTSIZE'] = subprocess.getoutput('du -csh /data1/input| tail -1 | cut -f1')
    os.environ['TEMPSIZE'] = subprocess.getoutput('du -csh /data1/tmp*| tail -1 | cut -f1')
    os.environ['OUTPUTSIZE'] = subprocess.getoutput('du -csh /data1/out| tail -1 | cut -f1')


def update_postrun_json_final(json_old, json_new, logfile=None):
    """Update postrun json with status, time stamps, parsed commands,
    input/tmp/output sizes"""
    prj = read_postrun_json(json_old)

    postrun_json_final(prj, logfile=logfile)

    # write to new json file
    write_postrun_json(json_new, prj)


def postrun_json_final(prj, logfile=None):
    # add commands
    if logfile:
        print("parsing commands from log file...")
        log_content = log.read_logfile_by_line(logfile)
        prj.update(commands=log.parse_commands(log_content))
    # add end time, status, instance_id
    prj_job = prj.Job
    prj_job.update(end_time=time.strftime("%Y%m%d-%H:%M:%S-%Z"))
    prj_job.update(status=os.getenv('JOB_STATUS'))
    prj_job.update(total_input_size=os.getenv('INPUTSIZE'))
    prj_job.update(total_tmp_size=os.getenv('TEMPSIZE'))
    prj_job.update(total_output_size=os.getenv('OUTPUTSIZE'))


def upload_postrun_json(jsonfile):
    prj = read_postrun_json(jsonfile)
    bucket = prj.Job.Log.log_bucket_directory
    dest = prj.Job.JOBID + '.postrun.json'
    if '/' in bucket:
        bucket_dirs = bucket.split('/')
        bucket = bucket_dirs.pop(0)
        prefix = '/'.join(bucket_dirs)
        dest = prefix + '/' + dest
    if prj.config.public_postrun_json:
        acl = 'public-read'
    else:
        acl = 'private'
    s3 = boto3.client('s3')
    upload_arg = {
        "Body": format_postrun_json(prj).encode('utf-8'),
        "Bucket": bucket,
        "Key": dest,
        "ACL": acl
    }
    if prj.config.encrypt_s3_upload:
        upload_arg.update({"ServerSideEncryption": "aws:kms"})
        if prj.config.kms_key_id:
            upload_arg['SSEKMSKeyId'] = prj.config.kms_key_id
    s3.put_object(**upload_arg)



"""
CLI for tibanna awsf3 package
"""

# -*- coding: utf-8 -*-
import argparse
import inspect
from tibanna._version import __version__  # for now use the same version as tibanna
from . import utils


PACKAGE_NAME = 'awsf3'


class Subcommands(object):

    def __init__(self):
        pass

    @property
    def descriptions(self):
        return {
            'decode_run_json': 'decode run json',
            'download_workflow': 'download workflow files',
            'update_postrun_json_init': 'update json json with instance ID and file system',
            'upload_postrun_json': 'upload postrun json file',
            'update_postrun_json_upload_output': 'update json json with output paths/target/md5 and upload outupt',
            'update_postrun_json_final': 'update postrun json with status, time stamp etc'
        }

    @property
    def args(self):
        return {
            'decode_run_json':
                [{'flag': ["-i", "--input-run-json"], 'help': "input run json file"},
                 {'flag': ["-k", "--kms-key-id"], 'help': "kms-key-id to use for encrypting s3 files"}],
            'download_workflow':
                [],
            'update_postrun_json_init':
                [{'flag': ["-i", "--input-json"], 'help': "input run/postrun json file"},
                 {'flag': ["-o", "--output-json"], 'help': "output postrun json file"}],
            'update_postrun_json_upload_output':
                [{'flag': ["-i", "--input-json"], 'help': "input run/postrun json file"},
                 {'flag': ["-e", "--execution-metadata-file"],
                  'help': "execution metadata file (output json of cwltool / cromwell)"},
                 {'flag': ["-m", "--md5file"], 'help': "text file storing md5 values for output files"},
                 {'flag': ["-o", "--output-json"], 'help': "output postrun json file"},
                 {'flag': ["-L", "--language"], 'help': "language", 'default': "cwl_v1"},
                 {'flag': ["-u", "--endpoint-url"], 'help': "s3 vpc endpoint url"}],
            'upload_postrun_json':
                [{'flag': ["-i", "--input-json"], 'help': "input postrun json file to upload to s3"}],
            'update_postrun_json_final':
                [{'flag': ["-i", "--input-json"], 'help': "input run/postrun json file"},
                 {'flag': ["-o", "--output-json"], 'help': "output postrun json file"},
                 {'flag': ["-l", "--logfile"], 'help': "Tibanna awsem log file"}],
        }


def decode_run_json(input_run_json, kms_key_id=None):
    utils.decode_run_json(input_run_json, kms_key_id=kms_key_id)


def download_workflow():
    utils.download_workflow()


def update_postrun_json_init(input_json, output_json):
    utils.update_postrun_json_init(input_json, output_json)


def update_postrun_json_upload_output(input_json, execution_metadata_file, md5file, output_json, language, endpoint_url):
    utils.update_postrun_json_upload_output(input_json, execution_metadata_file, md5file, output_json, language, endpoint_url=endpoint_url)


def upload_postrun_json(input_json):
    utils.upload_postrun_json(input_json)


def update_postrun_json_final(input_json, output_json, logfile):
    utils.update_postrun_json_final(input_json, output_json, logfile)


def main(Subcommands=Subcommands):
    """
    Execute the program from the command line
    """
    scs = Subcommands()

    # the primary parser is used for awsf -v or -h
    primary_parser = argparse.ArgumentParser(prog=PACKAGE_NAME, add_help=False)
    primary_parser.add_argument('-v', '--version', action='version',
                                version='%(prog)s ' + __version__)
    # the secondary parser is used for the specific run mode
    secondary_parser = argparse.ArgumentParser(prog=PACKAGE_NAME, parents=[primary_parser])
    subparsers = secondary_parser.add_subparsers(
        title=PACKAGE_NAME + ' subcommands',
        description='choose one of the following subcommands to run ' + PACKAGE_NAME,
        dest='subcommand',
        metavar='subcommand: {%s}' % ', '.join(scs.descriptions.keys())
    )
    subparsers.required = True

    def add_arg(name, flag, **kwargs):
        subparser[name].add_argument(flag[0], flag[1], **kwargs)

    def add_args(name, argdictlist):
        for argdict in argdictlist:
            add_arg(name, **argdict)

    subparser = dict()
    for sc, desc in scs.descriptions.items():
        subparser[sc] = subparsers.add_parser(sc, help=desc, description=desc)
        if sc in scs.args:
            add_args(sc, scs.args[sc])

    # two step argument parsing
    # first check for top level -v or -h (i.e. `tibanna -v`)
    (primary_namespace, remaining) = primary_parser.parse_known_args()
    # get subcommand-specific args
    args = secondary_parser.parse_args(args=remaining, namespace=primary_namespace)
    subcommandf = eval(args.subcommand)
    sc_args = [getattr(args, sc_arg) for sc_arg in inspect.getfullargspec(subcommandf).args]
    # run subcommand
    subcommandf(*sc_args)


if __name__ == '__main__':
    main()



def read_logfile_by_line(logfile):
    """generator function that yields the log file content line by line"""
    with open(logfile, 'r') as f:
        for line in f:
            yield line
    yield None


def parse_commands(log_content):
    """ 
    parse cwl commands from the line-by-line generator of log file content and 
    returns the commands as a list of command line lists, each corresponding to a step run.
    """ 
    command_list = []
    command = []
    in_command = False
    line = next(log_content)
    while(line):
        line = line.strip('\n')
        if '[job' in line and line.endswith('docker \\'):
            line = 'docker \\'  # remove the other stuff
            in_command = True
        if in_command:
            command.append(line.strip('\\').rstrip(' ')) 
            if not line.endswith('\\'):
                in_command = False
                command_list.append(command)
                command = []
        line = next(log_content)
    return(command_list)



import boto3
from . import create_logger


logger = create_logger(__name__)


def item2dict(item):
    '''convert a dynamoDB-style item to a regular dictionary'''
    return {k: list(v.values())[0] for k, v in item.items()}


def does_dynamo_table_exist(tablename):
    try:
        res = boto3.client('dynamodb').describe_table(
            TableName=tablename
        )
        if res:
            return True
        else:
            raise Exception("error describing table %s" % tablename)
    except Exception as e:
        if 'Requested resource not found' in str(e):
            return False
        else:
            raise Exception("error describing table %s" % tablename)


def create_dynamo_table(tablename, keyname):
    if does_dynamo_table_exist(tablename):
        logger.info("dynamodb table %s already exists. skip creating db" % tablename)
    else:
        response = boto3.client('dynamodb').create_table(
            TableName=tablename,
            AttributeDefinitions=[
                {
                     'AttributeName': keyname,
                     'AttributeType': 'S'
                }
            ],
            KeySchema=[
                {
                    'AttributeName': keyname,
                    'KeyType': 'HASH'
                 }
            ],
            BillingMode='PAY_PER_REQUEST'
        )


def get_items(table_name, primary_key, filter_key, filter_value, additional_keys=None):
    '''filter by filter_key=filter_value
    return all the values of primary_key and additional_keys.
    in the format of a list of dictionaries containing
    primary_key: value1, additional_key1: value2, additional_key2: value3, ...'''
    dd = boto3.client('dynamodb')

    if not additional_keys:
        additional_keys = []

    start_key = ''
    scan_input = {
        'TableName': table_name,
        'AttributesToGet': [primary_key] + additional_keys,
        'ScanFilter': {filter_key: {'AttributeValueList': [{'S': filter_value}],
                                    'ComparisonOperator': 'EQ'}}
    }
    entries = []

    while(True):
        if start_key:
            scan_input.update({'ExclusiveStartKey': {primary_key: {'S': start_key}}})
        res = dd.scan(**scan_input)
        if 'Items' in res and len(res['Items']) > 0:
            for item in res['Items']:
                entry = {primary_key: item[primary_key]['S']}
                for ak in additional_keys:
                    entry.update({ak: item[ak]['S']})
                entries.append(entry)
        if 'LastEvaluatedKey' in res and res['LastEvaluatedKey']:
            start_key = res['LastEvaluatedKey'][primary_key]['S']
        else:
            break
    return entries


def delete_items(table_name, primary_key, item_list, verbose=True):
    '''item_list is a list of dictionaries in the format of
    key1: value1, key2: value2, ...
    there has to be a primary key always.'''
    dd = boto3.client('dynamodb')
    for item in item_list:
        res2 = dd.delete_item(
            TableName=table_name,
            Key={primary_key: {'S': item[primary_key]}}
        )
    if verbose:
        logger.info("%d entries deleted from dynamodb." % len(item_list))



def combine_two(a, b, delimiter='/'):
    """returns an n-nested array of strings a+delimiter+b
    a and b (e.g. uuids and object_keys) can be a singlet,
    an array, an array of arrays or an array of arrays of arrays ...
    example:
    >>> a = ['a','b',['c','d']]
    >>> b = ['e','f',['g','h']]
    >>> combine_two(a, b)
    ['a/e','b/f',['c/g','d/h']]
    """
    if isinstance(a, list):
        if not isinstance(b, list):
            raise Exception("can't combine list and non-list")
        if len(a) != len(b):
            raise Exception("Can't combine lists of different lengths")
        return [combine_two(a_, b_) for a_, b_, in zip(a, b)]
    else:
        return(str(a) + delimiter + str(b))


def run_on_nested_arrays1(a, func, **param):
    """run func on each pair of element in a and b:
    a and b can be singlets, an array, an array of arrays, an array of arrays of arrays ...
    The return value is a flattened array
    """
    if isinstance(a, list):
        return([run_on_nested_arrays1(_, func, **param) for _ in a])
    else:
        return(func(a, **param))


def run_on_nested_arrays2(a, b, func, **param):
    """run func on each pair of element in a and b:
    a and b can be singlets, an array, an array of arrays, an array of arrays of arrays ...
    The return value is a flattened array
    """
    if isinstance(a, list):
        if not isinstance(b, list):
            raise Exception("can't combine list and non-list")
        if len(a) != len(b):
            raise Exception("Can't combine lists of different lengths")
        return([run_on_nested_arrays2(a_, b_, func, **param) for a_, b_ in zip(a, b)])
    else:
        return(func(a, b, **param))


def create_dim(a, dim='', empty=False):
    """create dimension array for n-nested array
    example:
    >>> create_dim([[1,2],[3,4],[5,6,[7,8],]])
    [['0-0', '0-1'], ['1-0', '1-1'], ['2-0', '2-1', ['2-2-0', '2-2-1']]]
    >>> create_dim(5)
    ''
    >>> create_dim([5,5])
    ['0', '1']
    >>> create_dim([5,5], empty=True)
    ['', '']
    """
    if isinstance(a, list):
        if dim:
            prefix = dim + '-'
        else:
            prefix = ''
        if empty:
            return([create_dim(a_, empty=empty) for a_ in a])
        else:
            return([create_dim(a_, prefix + str(i)) for i, a_ in enumerate(a)])
    else:
        return(dim)


def flatten(a):
    """recursively flatten n-nested array
    example:
    >>> flatten([[1,2],[3,4],[5,6,[7,8],]])
    [1, 2, 3, 4, 5, 6, 7, 8]
    """
    if isinstance(a, list):
        b = list()
        for a_ in a:
            if isinstance(a_, list):
                b.extend(flatten(a_))
            else:
                b.append(a_)
        return(b)
    else:
        return(a)



import re
import copy
from datetime import datetime
from .base import SerializableObject
from .ec2_utils import Config
from .exceptions import MalFormattedPostRunJsonException, MalFormattedRunJsonException
from .vars import AWSEM_TIME_STAMP_FORMAT
from .nnested_array import flatten


class AwsemRunJson(SerializableObject):
    def __init__(self, Job=None, config=None, strict=True, **kwargs):
        if strict:
            if not Job or not config:
                raise MalFormattedPostRunJsonException("Job and config are required fields.")
        if not Job:
            Job = {}
        self.create_Job(Job, strict=strict)
        self.config = Config(**config) if config else None

    def create_Job(self, Job, strict=True):
        self.Job = AwsemRunJsonJob(**Job, strict=strict)


class AwsemRunJsonJob(SerializableObject):
    def __init__(self, App=None, Input=None, Output=None, JOBID='',
                 start_time=None, Log=None, strict=True, **kwargs):
        if strict:
            if App is None or Input is None or Output is None or not JOBID:
                raise MalFormattedRunJsonException
        if not App:
            App = {}
        self.App = AwsemRunJsonApp(**App)
        if not Input:
            Input = {}
        self.Input = AwsemRunJsonInput(**Input)
        if not Output:
            Output = {}
        self.create_Output(Output)
        self.start_time = start_time
        self.JOBID = JOBID
        if not Log:
            Log = {}
        self.Log = AwsemRunJsonLog(**Log)

        # format check
        if self.App:
            self.Input.check_input_files_key_compatibility(self.App.language)

    def create_Output(self, Output):
        self.Output = AwsemPostRunJsonOutput(**Output)

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def start_time_as_datetime(self):
        if not self.start_time:
            return None
        else:
            return datetime.strptime(self.start_time, AWSEM_TIME_STAMP_FORMAT)


class AwsemRunJsonLog(SerializableObject):
    def __init__(self, log_bucket_directory=None, **kwargs):
        self.log_bucket_directory = log_bucket_directory


class AwsemRunJsonApp(SerializableObject):
    def __init__(self, App_name=None, App_version=None, language=None,
                 cwl_url=None, main_cwl=None, other_cwl_files=None,
                 wdl_url=None, main_wdl=None, other_wdl_files=None, workflow_engine=None, run_args=None,
                 container_image=None, command=None,
                 snakemake_url=None, main_snakemake=None, other_snakemake_files=None,
                 **kwargs):
        self.App_name = App_name
        self.App_version = App_version
        self.language = language
        self.cwl_url = cwl_url
        self.main_cwl = main_cwl
        self.other_cwl_files = other_cwl_files
        self.wdl_url = wdl_url
        self.main_wdl = main_wdl
        self.other_wdl_files = other_wdl_files
        self.workflow_engine = workflow_engine
        self.run_args = run_args
        self.container_image = container_image
        self.command = command
        self.snakemake_url = snakemake_url
        self.main_snakemake = main_snakemake
        self.other_snakemake_files = other_snakemake_files


class AwsemRunJsonInput(SerializableObject):
    def __init__(self, Input_files_data=None, Input_parameters=None, Secondary_files_data=None,
                 # Input_files_reference is for older postrunjson
                 # Env is missing in older postrunjson
                 Input_files_reference=None, Env=None, **kwargs):
        if not Input_files_data:
            Input_files_data = {}
        if not Input_parameters:
            Input_parameters = {}
        if not Secondary_files_data:
            Secondary_files_data = {}
        self.Input_files_data = {k: AwsemRunJsonInputFile(**v) for k, v in Input_files_data.items()}
        self.Secondary_files_data = {k: AwsemRunJsonInputFile(**v) for k, v in Secondary_files_data.items()}
        self.Input_parameters = Input_parameters
        self.Env = Env
        if Input_files_reference:
            self.Input_files_reference = {k: AwsemRunJsonInputFile(**v) for k, v in Input_files_reference.items()}

    def as_dict_as_cwl_input(self, input_dir='', input_mount_dir_prefix=''):
        d = {k: v.as_dict_as_cwl_input(input_dir, input_mount_dir_prefix) for k, v in self.Input_files_data.items()}
        d.update(self.Input_parameters)
        return d

    def as_dict_as_wdl_input(self, input_dir='', input_mount_dir_prefix=''):
        d = {k: v.as_dict_as_wdl_input(input_dir, input_mount_dir_prefix) for k, v in self.Input_files_data.items()}
        d.update(self.Input_parameters)
        return d

    def check_input_files_key_compatibility(self, language):
        for category in ["Input_files_data", "Secondary_files_data"]:
            for inkey in getattr(self, category):
                if inkey.startswith('file://'):
                    if language not in ['shell', 'snakemake']:
                        raise MalFormattedRunJsonException('input file has to be defined with argument name for CWL and WDL')
                    target = inkey.replace('file://', '') 
                    if not target.startswith('/data1/'):
                        raise Exception('input target directory must be in /data1/')
                    if not target.startswith('/data1/' + language) and \
                        not target.startswith('/data1/input') and \
                        not target.startswith('/data1/out'):
                            raise Exception('input target directory must be in /data1/input, /data1/out or /data1/%s' % language)


class AwsemRunJsonInputFile(SerializableObject):
    def __init__(self, path, profile='', rename='', unzip='', mount=False, **kwargs):  # kwargs includes 'dir' and 'class'
        # profile and rename are missing in the old postrunjson
        self.path = path
        self.profile = profile
        self.rename = rename
        self.unzip = unzip
        self.mount = mount
        # handling reserved name key
        self.class_ = kwargs.get('class', None)
        self.dir_ = kwargs.get('dir', None)

        # field compatibility check
        errmsg = "Incompatible input for file %s: %s and mount cannot be used together."
        if self.rename and len(flatten(self.rename))>0 and self.mount:  # the second condition covers e.g. []
            raise MalFormattedRunJsonException(errmsg % (self.path, 'rename'))
        if self.unzip and self.mount:
            raise MalFormattedRunJsonException(errmsg % (self.path, 'unzip'))

    def as_dict(self):
        d = super().as_dict()
        # handling reserved name key
        for rk in ['class', 'dir']:
            rk_alt = rk + '_'
            if rk_alt in d:
                d[rk] = d[rk_alt]
                del(d[rk_alt])
        return d

    def as_dict_as_cwl_input(self, input_dir='', input_mount_dir_prefix=''):
        if self.mount:
            input_dir = input_mount_dir_prefix + self.dir_
        else:
            input_dir = input_dir
        if self.rename:
            if isinstance(self.rename, list):
                path = self.rename[:]
            else:
                path = self.rename
        else:
            path = self.path
        if isinstance(path, list):
            d = []
            for pi in path:
                if isinstance(pi, list):
                    nested = []
                    for ppi in pi:
                        if isinstance(ppi, list):
                            nested.append([file2cwlfile(pppi, input_dir, self.unzip) for pppi in ppi])
                        else:
                            nested.append(file2cwlfile(ppi, input_dir, self.unzip))
                    d.append(nested)
                else:
                    d.append(file2cwlfile(pi, input_dir, self.unzip))
            return d
        else:
            return file2cwlfile(path, input_dir, self.unzip)

    def as_dict_as_wdl_input(self, input_dir='', input_mount_dir_prefix=''):
        if self.mount:
            input_dir = input_mount_dir_prefix + self.dir_
        else:
            input_dir = input_dir
        if self.rename:
            if isinstance(self.rename, list):
                path = list(self.rename)
            else:
                path = self.rename
        else:
            path = self.path
        if isinstance(path, list):
            d = []
            for pi in path:
                if isinstance(pi, list):
                    nested = []
                    for ppi in pi: 
                        if isinstance(ppi, list):
                            nested.append([file2wdlfile(pppi, input_dir, self.unzip) for pppi in ppi])
                        else:
                            nested.append(file2wdlfile(ppi, input_dir, self.unzip))
                    d.append(nested)
                else:
                    d.append(file2wdlfile(pi, input_dir, self.unzip))
            return d
        else:
            return file2wdlfile(path, input_dir, self.unzip)


class AwsemRunJsonOutput(SerializableObject):
    def __init__(self, output_bucket_directory=None, output_target=None,
                 secondary_output_target=None, alt_cond_output_argnames=None, **kwargs):
        self.output_bucket_directory = output_bucket_directory or {}
        self.output_target = output_target or {}
        self.secondary_output_target = secondary_output_target or {}
        self.alt_cond_output_argnames = alt_cond_output_argnames or {}

        for u, v in self.secondary_output_target.items():
            if not isinstance(v, list):
                self.secondary_output_target[u] = [v]

    def alt_output_target(self, argname_list):
        """In case conditional alternative output targets exist, return alternative output target
        where the output target keys are replaced with the alternative names.
        If not, return output_target itself.
        This function does not actually modify output_target.
        It cannot be applied to custom output targets starting with 'file://'
        We don't need to do the same for secondary files because
        conditional alternative names only occur in WDL which does not support secondary files"""

        # first create a list of keys to be replaced
        replace_list = []
        for k in self.output_target:
            if k.startswith('file://'):
                continue
            if k not in argname_list:
                if k in self.alt_cond_output_argnames:
                    key_exists = False  # initialize
                    for k_alt in self.alt_cond_output_argnames[k]:
                        if k_alt in argname_list:
                            key_exists = True
                            replace_list.append((k, k_alt))
                            break
                    if not key_exists:
                        raise Exception("output target key %s doesn't exist in argname list" % k)
                else:
                    raise Exception("output target key %s doesn't exist in argname list" % k)

        # return the alternated output_target
        alt_output_target = copy.deepcopy(self.output_target)
        for k, k_alt in replace_list:
            alt_output_target[k_alt] = alt_output_target[k]
            del alt_output_target[k]
        return alt_output_target


class AwsemPostRunJson(AwsemRunJson):
    def __init__(self, Job=None, config=None, commands=None, log=None, strict=True, **kwargs):
        if strict:
            if not Job or not config:
                raise MalFormattedPostRunJsonException("Job and config are required fields.")
        super().__init__(Job, config, strict=strict)
        if commands:
            self.commands = commands
        if log:
            self.log = log

    def add_commands(self, command):
        self.command = command

    def create_Job(self, Job, strict=True):
        self.Job = AwsemPostRunJsonJob(**Job, strict=strict)


class AwsemPostRunJsonJob(AwsemRunJsonJob):
    def __init__(self, App=None, Input=None, Output=None, JOBID='',
                 start_time=None, end_time=None, status=None, Log=None,
                 total_input_size=None, total_output_size=None, total_tmp_size=None,
                 # older postrunjsons don't have these fields
                 filesystem='', instance_id='', instance_availablity_zone='', instance_type='',
                 Metrics=None, strict=True, **kwargs):
        if strict:
            if App is None or Input is None or Output is None or not JOBID or start_time is None:
                errmsg = "App, Input, Output, JOBID and start_time are required fields"
                raise MalFormattedPostRunJsonException(errmsg)
        super().__init__(App, Input, Output, JOBID, start_time, Log, strict=strict)
        self.end_time = end_time
        self.status = status
        self.filesystem = filesystem
        self.instance_id = instance_id
        self.instance_availablity_zone = instance_availablity_zone
        self.instance_type = instance_type
        self.total_input_size = total_input_size
        self.total_output_size = total_output_size
        self.total_tmp_size = total_tmp_size
        self.Metrics = Metrics

    def create_Output(self, Output):
        self.Output = AwsemPostRunJsonOutput(**Output)

    @property
    def end_time_as_datetime(self):
        try:
            return datetime.strptime(self.end_time, AWSEM_TIME_STAMP_FORMAT)
        except:
            return None

    def add_filesystem(self, filesystem):
        self.filesystem = filesystem


class AwsemPostRunJsonOutput(AwsemRunJsonOutput):
    def __init__(self, output_bucket_directory=None, output_target=None,
                 secondary_output_target=None, alt_cond_output_argnames=None,
                 **kwargs):  # kwargs includes 'Output files'
        """This class has an additional 'Output files' field which
        stores the output from CWL/WDL runs"""
        super().__init__(output_bucket_directory, output_target,
                         secondary_output_target, alt_cond_output_argnames)
        if 'Output files' in kwargs:
            self.add_output_files(kwargs['Output files'])
        else:
            self.Output_files_ = {}

    @property
    def output_files(self):
        return self.Output_files_

    def add_output_files(self, output_files):
        """add or replace output files. output_files is a dictionary with argnames as keys
           and a dict form of AwsemPostRunJsonOutputFile objects as values"""
        self.Output_files_ = {k: AwsemPostRunJsonOutputFile(**v) for k, v in output_files.items()}

    def as_dict(self):
        d = super().as_dict()
        # handling reserved name key
        if 'Output_files_' in d:
            d['Output files'] = d['Output_files_']
            del(d['Output_files_'])
        return d


class AwsemPostRunJsonOutputFile(SerializableObject):
    def __init__(self, path, target=None, basename=None, checksum=None,
                 location=None, md5sum=None, size=None, secondaryFiles=None,
                 **kwargs):  # kwargs includes 'class'
        # both WDL and CWL
        self.path = path
        self.target = target
        # currently CWL-only
        self.basename = basename
        self.checksum = checksum
        self.location = location
        self.md5sum = md5sum
        self.size = size
        if secondaryFiles:
            if isinstance(secondaryFiles, list):
                self.secondaryFiles = [AwsemPostRunJsonOutputFile(**sf) for sf in secondaryFiles]
            else:
                raise MalFormattedPostRunJsonException("secondaryFiles must be a list")
        else:
            self.secondaryFiles = None
        # handling reserved name key
        self.class_ = kwargs.get('class', None)

    def add_target(self, target):
        self.target = target

    def as_dict(self):
        d = super().as_dict()
        # handling reserved name key
        if 'class_' in d:
            d['class'] = d['class_']
            del(d['class_'])
        return d


def file2cwlfile(filename, dirname, unzip):
    if unzip:
        filename = re.match('(.+)\.{0}$'.format(unzip), filename).group(1)
    if dirname.endswith('/'):
        dirname = dirname.rstrip('/')
    return {"class": 'File', "path": dirname + '/' + filename}


def file2wdlfile(filename, dirname, unzip):
    if unzip:
        filename = re.match('(.+)\.{0}$'.format(unzip), filename).group(1)
    if dirname.endswith('/'):
        dirname = dirname.rstrip('/')
    return dirname + '/' + filename



import os
import boto3
import sys
from datetime import datetime
from dateutil.tz import tzutc
from ._version import __version__
from . import create_logger


logger = create_logger(__name__)


if boto3.session.Session().get_credentials() is None:
    logger.info('Please provide AWS credentials.')
    sys.exit(-1)


# AWS account info
AWS_ACCOUNT_NUMBER = os.environ.get('AWS_ACCOUNT_NUMBER', '')
if not AWS_ACCOUNT_NUMBER:
    try:
        AWS_ACCOUNT_NUMBER = boto3.client('sts').get_caller_identity().get('Account')
    except Exception as e:
        raise Exception("Cannot obtain AWS account number. Please provide AWS credentials")

AWS_REGION = os.environ.get('TIBANNA_AWS_REGION', '')
if not AWS_REGION:
    # I'm a lambda
    AWS_REGION = os.environ.get('AWS_REGION', '')  # reserved variable in lambda
    # I'm a user
    if not AWS_REGION:
        try:
            AWS_REGION = boto3.session.Session().region_name  # for a user
        except Exception as e:
            raise Exception("Cannot find AWS_REGION: %s" % e)


# Tibanna AMI info
# Override this mapping to use a custom AMI scheme
AMI_PER_REGION = {
    'x86': {
        'us-east-1': 'ami-06e2266f85063aabc',  # latest as of Oct 25 2021
        'us-east-2': 'ami-03a4e3e84b6a1813d',
        'us-west-1': 'ami-0c5e8147be760a354',
        'us-west-2': 'ami-068589fed9c8d5950',
        'ap-south-1': 'ami-05ef59bc4f359c93b',
        'ap-northeast-2': 'ami-0d8618a76aece8a8e',
        'ap-southeast-1': 'ami-0c22dc3b05714bda1',
        'ap-southeast-2': 'ami-03dc109bbf412aac5',
        'ap-northeast-1': 'ami-0f4c520515c41ff46',
        'ca-central-1': 'ami-01af127710fadfe74',
        'eu-central-1': 'ami-0887bcb1c901c1769',
        'eu-west-1': 'ami-08db59692e4371ea6',
        'eu-west-2': 'ami-036d3ce7a21e07012',
        'eu-west-3': 'ami-0cad0ec4160a6b940',
        'eu-north-1': 'ami-00a6f0f9fee951aa0',
        'sa-east-1': 'ami-0b2164f9680f97099',
        'me-south-1': 'ami-03479b7a590f97945',
        'af-south-1': 'ami-080baa4ec59c456aa',
        'ap-east-1': 'ami-0a9056eb817bc3928',
        'eu-south-1': 'ami-0a72279e56849415e'
    },
    'Arm': {
        'us-east-1': 'ami-0f3e90ad8e76c7a32', # latest as of Nov 23 2022
        'us-east-2': 'ami-03359d89f311a015e', 
        'us-west-1': 'ami-00ffd20b39dbfb6ea', 
        'us-west-2': 'ami-08ab3015c1bc36d24', 
        'ap-south-1': 'ami-01af9ec07fed38a38', 
        'ap-northeast-2': 'ami-0ee2af459355dd917', 
        'ap-southeast-1': 'ami-0d74dc5af4bf74ea8', 
        'ap-southeast-2': 'ami-08ab7201c83209fe8', 
        'ap-northeast-1': 'ami-07227003bfa0565e3', 
        'ca-central-1': 'ami-0cbf87c80362a058e', 
        'eu-central-1': 'ami-09cfa59f75e88ad54', 
        'eu-west-1': 'ami-0804bdeafd8af01f8', 
        'eu-west-2': 'ami-0db05a333dc02c1c8', 
        'eu-west-3': 'ami-0ceab436f882fe36a', 
        'eu-north-1': 'ami-04ba962c974ddd374', 
        'sa-east-1': 'ami-0fc9a9dec0f3df318', 
        'me-south-1': 'ami-0211bc858eb163594', 
        'af-south-1': 'ami-0d6a4af087f83899d', 
        'ap-east-1': 'ami-0d375f2ce688d16b9', 
        'eu-south-1': 'ami-0b1db84f31597a70f'
    }
}

AWS_REGION_NAMES = {
    'us-east-1': 'US East (N. Virginia)',
    'us-east-2': 'US East (Ohio)',
    'us-west-1': 'US West (N. California)',
    'us-west-2': 'US West (Oregon)',
    'ca-central-1': 'Canada (Central)',
    'eu-north-1': 'EU (Stockholm)',
    'eu-west-1': 'EU (Ireland)',
    'eu-central-1': 'EU (Frankfurt)',
    'eu-west-2': 'EU (London)',
    'eu-west-3': 'EU (Paris)',
    'ap-northeast-1': 'Asia Pacific (Tokyo)',
    'ap-northeast-2': 'Asia Pacific (Seoul)',
    'ap-northeast-3': 'Asia Pacific (Osaka-Local)',
    'ap-southeast-1': 'Asia Pacific (Singapore)',
    'ap-southeast-2': 'Asia Pacific (Sydney)',
    'ap-south-1': 'Asia Pacific (Mumbai)',
    'sa-east-1': 'South America (Sao Paulo)',  # intentionally no unicode,
    'us-gov-west-1': 'AWS GovCloud (US)',
    'us-gov-east-1': 'AWS GovCloud (US-East)'
}

# Tibanna repo from which awsf scripts are pulled
TIBANNA_REPO_NAME = os.environ.get('TIBANNA_REPO_NAME', '4dn-dcic/tibanna')
TIBANNA_REPO_BRANCH = os.environ.get('TIBANNA_REPO_BRANCH', 'master')
TIBANNA_AWSF_DIR = 'awsf3'

# Tibanna roles
AWS_S3_ROLE_NAME = os.environ.get('AWS_S3_ROLE_NAME', 'S3_access')
S3_ACCESS_ARN = 'arn:aws:iam::' + AWS_ACCOUNT_NUMBER + ':instance-profile/' + AWS_S3_ROLE_NAME

# Profile keys (optional) to use on AWSEM EC2
TIBANNA_PROFILE_ACCESS_KEY = os.environ.get('TIBANNA_PROFILE_ACCESS_KEY', '')
TIBANNA_PROFILE_SECRET_KEY = os.environ.get('TIBANNA_PROFILE_SECRET_KEY', '')

# default step function name
TIBANNA_DEFAULT_STEP_FUNCTION_NAME = os.environ.get('TIBANNA_DEFAULT_STEP_FUNCTION_NAME', 'tibanna_unicorn')

# S3_ENCRYPT_KEY_ID for Tibanna output buckets
S3_ENCRYT_KEY_ID = os.environ.get('S3_ENCRYPT_KEY_ID', None)

# dynamo table (optional) for fast searching
DYNAMODB_TABLE = 'tibanna-master'
DYNAMODB_KEYNAME = 'Job Id'

# field name reserved for Tibanna setting
_tibanna = '_tibanna'

# Awsem time stamp format
AWSEM_TIME_STAMP_FORMAT = '%Y%m%d-%H:%M:%S-UTC'


def PARSE_AWSEM_TIME(t_str):
  t = datetime.strptime(t_str, AWSEM_TIME_STAMP_FORMAT)
  return t.replace(tzinfo=tzutc())


# EBS mount path for cloudwatch metric collection
EBS_MOUNT_POINT = '/mnt/data1'

# Default root EBS size
DEFAULT_ROOT_EBS_SIZE = 8

# Default awsf image
DEFAULT_AWSF_IMAGE = '4dndcic/tibanna-awsf:' + __version__

SFN_TYPE = 'unicorn'
LAMBDA_TYPE = ''
RUN_TASK_LAMBDA_NAME = 'run_task_awsem'
CHECK_TASK_LAMBDA_NAME = 'check_task_awsem'
UPDATE_COST_LAMBDA_NAME = 'update_cost_awsem'

# step function and execution ARN generators
BASE_ARN = 'arn:aws:states:' + AWS_REGION + ':' + AWS_ACCOUNT_NUMBER + ':%s:%s'
BASE_EXEC_ARN = 'arn:aws:states:' + AWS_REGION + ':' + AWS_ACCOUNT_NUMBER + ':execution:%s:%s'
BASE_METRICS_URL = 'https://%s.s3.amazonaws.com/%s.metrics/metrics.html'

METRICS_COLLECTION_INTERVAL = 120 # in seconds, same value as in cloudwatch_agent_config.json


def STEP_FUNCTION_ARN(sfn=TIBANNA_DEFAULT_STEP_FUNCTION_NAME):
    return BASE_ARN % ('stateMachine', sfn)


def EXECUTION_ARN(exec_name, sfn=TIBANNA_DEFAULT_STEP_FUNCTION_NAME):
    return BASE_EXEC_ARN % (sfn, exec_name)


def METRICS_URL(log_bucket, job_id):
    return BASE_METRICS_URL % (log_bucket, job_id)
